{$R+}
{$Q+}
(* {$S-} *)

unit Unit1;

{ Program to calculate Pi using some of the amazing formulas recently found
  on the Web, including the "Miraculous" Bailey-Borwein-Plouffe
  algorithm. This simply verifies the formulas by doing them all in extended
  precision until they converges.
  Written by Chris Williams, 20/7/97. }

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ComCtrls, Menus;

type
  TForm1 = class(TForm)
    RichEdit1: TRichEdit;
    Label1: TLabel;
    MainMenu1: TMainMenu;
    Bailey1: TMenuItem;
    Allextendedcalculations1: TMenuItem;
    Allextendedwithfunctions1: TMenuItem;
    Mixedextendedinteger1: TMenuItem;
    HighPrecision1: TMenuItem;
    Ramanujan1: TMenuItem;
    AllExtendedCalculations2: TMenuItem;
    Chudnovsky1: TMenuItem;
    AllExtendedCalculations3: TMenuItem;
    Improvedwithstats1: TMenuItem;
    Algorithm11: TMenuItem;
    AllExtendedCalculations4: TMenuItem;
    procedure Allextendedcalculations1Click(Sender: TObject);
    procedure Allextendedwithfunctions1Click(Sender: TObject);
    procedure Mixedextendedinteger1Click(Sender: TObject);
    procedure HighPrecision1Click(Sender: TObject);
    procedure AllExtendedCalculations2Click(Sender: TObject);
    procedure AllExtendedCalculations3Click(Sender: TObject);
    procedure Improvedwithstats1Click(Sender: TObject);
    procedure AllExtendedCalculations4Click(Sender: TObject);
  private
    { Private declarations }
    procedure wr(s : string);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

uses supportx;

{$R *.DFM}

type
  TStringNum = array [0 .. 1000] of byte;
  Tprecise = extended;
  Tzprecise = extended;
  Tzlongint = extended;

var
  sum, last, n, m, t8 : Tprecise;
  zsum, zlast : Tzprecise;
  zm, zn, zt8 : Tzlongint;

procedure TForm1.wr(s : string);
{ Simplified way to display lines on the screen }
begin
  RichEdit1.Lines.Add(s)
end;

{ **************************************************************************** }

{ Here are the "extended" versions of the "high precision" primitive
  instructions. These are based on the idea that some speed gain /
  simplification may be possible with some arguments being small integer
  constants. Note that there is no "e_div", because all divisions have a
  small integer constant as either a divisor or dividend! }

function e_add(x1, x2 : Tprecise) : Tprecise;
begin
  e_add := x1+x2;
end;

function e_sub(x1, x2 : Tprecise) : Tprecise;
begin
  e_sub := x1-x2;
end;

function e_mul(x1, x2 : Tprecise) : Tprecise;
begin
  e_mul := x1*x2;
end;

function e_mul8(x : Tprecise) : Tprecise;
begin
  e_mul8 := x*8.0;
end;

function e_add1(x : Tprecise) : Tprecise;
begin
  e_add1 := x+1.0;
end;

function e_add4(x : Tprecise) : Tprecise;
begin
  e_add4 := x+4.0;
end;

function e_add5(x : Tprecise) : Tprecise;
begin
  e_add5 := x+5.0;
end;

function e_add6(x : Tprecise) : Tprecise;
begin
  e_add6 := x+6.0;
end;

function e_4divby(x : Tprecise) : Tprecise;
begin
  e_4divby := 4.0/x;
end;

function e_2divby(x : Tprecise) : Tprecise;
begin
  e_2divby := 2.0/x;
end;

function e_1divby(x : Tprecise) : Tprecise;
begin
  e_1divby := 1.0/x;
end;

function e_div16(x : Tprecise) : Tprecise;
begin
  e_div16 := x/16.0;
end;

{ Here are the other versions of the "high precision" primitive instructions.
  These are more basic arithmetic operations, taking into account those cases
  where integer types are used! Note that division is only ever used with 1, 2
  or 4 as the dividend (above line) and a long positive integer divisor (below
  line); OR - a series of x/m where m:=1, 16, 16*16, 16*16*16 etc. Note also
  that all results (including subtraction) are positive, so no sign
  representation is required in numbers! Also, all floating/fixed point results
  are in the range 0.0 .. 9.9, so we only need to cater for one decimal digit
  to the left of the point, and an arbitary number to the right! }

{ More work required below: these go out of range with some calculations! Once
  the guilty calls have been identified, the "i" functions can be converted
  to actually use longint }

procedure rx(x : Tzprecise);
{ Checks range of floating point values }
begin
  if (x < 0.0) or {(x > 9.9) or} (x > 1.0e99) then
    ShowMessage('FP value out of range 0..3.15: '+em(x,12));
end;

procedure ri(n : Tzlongint);
{ Checks range of integer values, and if is really is integer when
   Tzlongint is actually extended! }
begin
  if (n < 0) {or (n > 1000) or (pos('.',em(n,12)) <>0)} then
    ShowMessage('Int value not integer or out of range 0..1000: '+em(n,12));
end;

function z_add(x1, x2 : Tzprecise) : Tzprecise;
begin
  rx(x1); rx(x2); rx(x1+x2);
  z_add := x1+x2;
end;

function z_iadd(n1, n2 : Tzlongint) : Tzlongint;
begin
  ri(n1); ri(n2); ri(n1+n2);
  z_iadd := n1+n2;
end;

function z_sub(x1, x2 : Tzprecise) : Tzprecise;
begin
  rx(x1); rx(x2); rx(x1-x2);
  z_sub := x1-x2;
end;

function z_imul8(n : Tzlongint) : Tzlongint;
begin
  ri(n); ri(n*8);
  z_imul8 := n*8;
end;

function z_imul16(n : Tzlongint) : Tzlongint;
begin
  ri(n); ri(n*16);
  z_imul16 := n*16;
end;

function z_ndivx(n : Tzlongint; x : Tzprecise) : Tzprecise;
begin
  ri(n); rx(x); rx(n/x);
  z_ndivx := n/x;
end;

function z_xdivn(x : Tzprecise; n : Tzlongint) : Tzprecise;
begin
  rx(x); ri(n); rx(x/n);
  z_xdivn := x/n;
end;
{ **************************************************************************** }

procedure TForm1.Allextendedcalculations1Click(Sender: TObject);
begin
  { Calculate successive terms of Pi using the "miraculous"
    Bailey-Borwein-Plouffe algorithm }
  n := 0.0;
  m := 1.0;
  sum := 0.0;
  wr('Calculation of Pi using the "Miraculous" Bailey-Borwein-Plouffe algorithm.');
  wr('This uses straight extended precision arithmetic ...');
  repeat
    last := sum;
    sum := sum + ( (4.0/(8.0*n+1.0)) - (2.0/(8.0*n+4.0)) -
                   (1.0/(8.0*n+5.0)) - (1.0/(8.0*n+6.0)) ) * m;
    m := m/16.0;
    wr('Term '+em(n,6)+#9#9+em(sum,20));
    RichEdit1.Refresh;
    n := n + 1.0;
  until sum = last;
  wr('Difference between this value and the built-in "Pi" function = '+
     em(sum-Pi,20));
  wr(' ');
end;

procedure TForm1.Allextendedwithfunctions1Click(Sender: TObject);
begin
  { Ok, now we will "decompose" the formula into a series of function calls.
    For now, still use extended math, but this paves the way for string
    representation of more precise number at the next stage! }
  n := 0.0;
  m := 1.0;
  sum := 0.0;
  wr('Calculation of Pi using the "Miraculous" Bailey-Borwein-Plouffe algorithm.');
  wr('This uses a series of primitive functions ...');
  repeat
    last := sum;
    { Calculate repeated constant expression first }
    t8 := e_mul8(n);
    { Now call functions! }
    { sum := sum + ( (4.0/(8.0*n+1.0)) - (2.0/(8.0*n+4.0)) -
                   (1.0/(8.0*n+5.0)) - (1.0/(8.0*n+6.0)) ) * m; }
    sum := e_add(sum,e_mul(e_sub(e_sub(e_sub(
           e_4divby(e_add1(t8)),
           e_2divby(e_add4(t8))),
           e_1divby(e_add5(t8))),
           e_1divby(e_add6(t8))),m));
    { m := m/16.0; }
    m := e_div16(m);
    wr('Term '+em(n,6)+#9#9+em(sum,20));
    RichEdit1.Refresh;
    n := n + 1.0;
  until sum = last;
  wr('Difference between this value and the built-in "Pi" function = '+
     em(sum-Pi,20));
  wr(' ');
end;

procedure TForm1.Mixedextendedinteger1Click(Sender: TObject);
begin
  { Ok, next we will "decompose" the formula into just the four basic
    arithmetic functions.
    For now, still use extended math, but this paves the way for string
    representation of more precise number at the next stage! }
  zn:= 0;
  zm := 1;
  zsum := 0.0;
  wr('Calculation of Pi using the "Miraculous" Bailey-Borwein-Plouffe algorithm.');
  wr('This uses just the four basic arithmetic operators ...');
  repeat
    zlast := zsum;
    { Calculate repeated constant expression first }
    zt8 := z_imul8(zn);
    { Now call functions! }
    { zsum := zsum + ( (4.0/(8.0*n+1.0)) - (2.0/(8.0*n+4.0)) -
                   (1.0/(8.0*n+5.0)) - (1.0/(8.0*n+6.0)) ) / m; }
    zsum := z_add(zsum,z_xdivn(z_sub(z_sub(z_sub(
            z_ndivx(4,z_iadd(zt8,1)),
            z_ndivx(2,z_iadd(zt8,4))),
            z_ndivx(1,z_iadd(zt8,5))),
            z_ndivx(1,z_iadd(zt8,6))),zm));
    { m := m*16; }
    zm := z_imul16(zm);
    wr('Term '+em(zn,6)+#9#9+em(zsum,20));
    RichEdit1.Refresh;
    zn := zn + 1.0;
  until zsum = zlast;
  wr('Difference between this value and the built-in "Pi" function = '+
     em(zsum-Pi,20));
  wr(' ');
end;

procedure TForm1.HighPrecision1Click(Sender: TObject);

{ This is a completely new section which will use the very different, and
  vastly improved algorithm originally from the "pi_c.txt" file, coupled
  with my own code to do an extended precision hex to decimal convertion.
  Under construction - just a test shell at the moment. }

const
  MaxDigits = 50;

type
  { The Tnum type represents a decimal number. Each array element is set to
    a value of 0 to 9 to represent the digit }
  Tnum = array [1 .. MaxDigits] of byte;

function nadd (n1, n2 : Tnum) : Tnum;
{ Adds n1 to n2 and returns result }
var
  i : integer;
  t, rem : byte;
  temp : Tnum;
begin
  rem := 0;
  for i := MaxDigits downto 1 do
  begin
    t := n1[i] + n2[i] + rem;
    temp[i] := t mod 10;
    rem := t div 10
  end;
  nadd := temp
end;

function ndiv4 (n : Tnum) : Tnum;
{ Divides n by 4, discards remainder if any, and returns result }
var
  i : integer;
  t, rem : byte;
  temp : Tnum;
begin
  rem := 0;
  for i := 1 to MaxDigits do
  begin
    t := n[i] + (rem * 10);
    temp[i] := t div 4;
    rem := t mod 4;
  end;
  ndiv4 := temp
end;

function nconv (s : string) : Tnum;
{ Converts the decimal digits in string into a Tnum array of bytes }
var
  i : integer;
  t : Tnum;
begin
  fillchar (t, sizeof(t), 0);
  if length(s) > sizeof(t) then
  begin
    ShowMessage('String too long to fit into Tnum array!');
    SetLength(s, sizeof(t))
  end;
  for i := 1 to length(s) do
  begin
    if s[i] in ['0' .. '9'] then
      t[i+MaxDigits-Length(s)] := byte(s[i]) - 48;
  end;
  nconv := t
end;

function nout (t : Tnum; num : Integer) : String;
{ Converts the digits in a Tnum array of bytes to a displayable string. Only
   the first "num" digits are output.}
var
  i, limit : integer;
  s : String;
begin
  s := '';
  if num > sizeof(t) then
    limit := sizeof(t)
  else
    limit := num;
  for i := 1 to limit do
  begin
    if t[i] > 9 then
    begin
      ShowMessage('Invalid value ( >9 ) in Tnum variable passed to nout!');
      s := s + '?'
    end
    else
    begin
      s := s + char(t[i] + 48)
    end
  end;
  nout := s
end;

{ Test code }
var
  xdec, ydec : String;
  x, y : Tnum;

begin  { of TForm1.HighPrecision1Click }
  fillchar(x, sizeof(x), 0);
  fillchar(y, sizeof(y), 0);
  if InputQuery ('Test High Precision arithmetic functions',
                 'Give 2 decimal values separated by a space', ydec) then
  begin
    ydec := Trim(ydec);
    xdec := Trim(parse(ydec, ' '));
    if (xdec = '') or (ydec = '') then
      ShowMessage('Invalid input!')
    else
    begin
      x := nconv(xdec);
      y := nconv(ydec);
      wr('x dec=  '+#9+xdec);
      wr('y dec=  '+#9+ydec);
      wr('x:      '+#9+nout(x,MaxDigits));
      wr('y:      '+#9+nout(y,MaxDigits));
      wr('nadd:   '+#9+nout(nadd(x,y),MaxDigits));
      wr('ndiv4 x:'+#9+nout(ndiv4(x), MaxDigits));
      wr('ndiv4 y:'+#9+nout(ndiv4(y), MaxDigits));
      wr(' ');
    end
  end
  else
    ShowMessage('No input');
end;

procedure TForm1.AllExtendedCalculations2Click(Sender: TObject);
{ This uses the amazing Ramanujan Pi formula! }
function fact (n : extended) : extended;
{ A local factorial function, where the argument "n" is extended, and allows
  for possible rounding errors (e.g. if n = 3.9999999, it will be taken as 4) }
var
  t : extended;
begin
  if n <= 1.5 then
    fact := 1.0
  else
  begin
    t := n;
    while n > 1.5 do
    begin
      n := n - 1.0;
      t := t * n
    end;
    fact := t
  end
end;  { of fact }
function pwr4 (x : extended) : extended;
begin
  pwr4 := x*x*x*x
end;
const
  cpwr = 396.0 * 396.0 * 396.0 * 396.0;
var
  n, term, f1, f2, pi1, zpi, lastterm : extended;
  { This is a table to do the calculation of "396.0^(4*n)". It is limited
    to 200 because it soon overflows the range of extended :- }
  powertable : array [0 .. 200] of extended;
  i : integer;
begin  { of TForm1.AllExtendedCalculations2Click }
  wr('Calculation of Pi using the amazing Ramanujan formula');
  { Set up table for the term 396.0^(4*n) }
  powertable[0] := 1.0;
  for i := 1 to 200 do
    try
      powertable[i] := powertable[i-1]*cpwr;
    except
      ShowMessage('Failed on table fill at '+IntToStr(i));
      break
    end;
  term := 0.0;
  n := 0.0;
  wr('NOTE: Pi = approximately  1 / (0.00028858556*Term)');
  repeat
    lastterm := term;
    f1 := fact(4.0*n) / pwr4(fact(n));
    f2 := (1103.0 + 26390.0*n) / powertable[round(n)];
    term := term + f1*f2;
    n := n + 1.0;
    wr('Term '+em(n,6)+#9#9+em(term,20))
  until lastterm = term;
  pi1 := sqrt(8.0)/9801.0*term;
  zpi := 1.0 / pi1;
  wr('Converged to limit of extended precision after only '+em(n,12)+' terms!');
  wr('Pi = '+em(zpi,20));
  wr('Difference between this value and the built-in "Pi" function = '+
     em(zpi-Pi,20));
  wr(' ');
end;  { of TForm1.AllExtendedCalculations2Click }

procedure TForm1.AllExtendedCalculations3Click(Sender: TObject);
{ This uses the even more amazing Chudnovsky Brothers Pi formula! }
function fact (n : extended) : extended;
{ A local factorial function, where the argument "n" is extended, and allows
  for possible rounding errors (e.g. if n = 3.9999999, it will be taken as 4) }
var
  t : extended;
begin
  if n <= 1.5 then
    fact := 1.0
  else
  begin
    t := n;
    while n > 1.5 do
    begin
      n := n - 1.0;
      t := t * n
    end;
    fact := t
  end
end;  { of fact }
function pwr3 (x : extended) : extended;
begin
  pwr3 := x*x*x
end;
const
  const3 = 640320.0 * 640320.0 * 640320.0;
var
  n, term, flip, f1, f2, pi1, zpi, lastterm : extended;
  { This is a table to do the calculation of "(640320^3)^n". It is limited
    to .. because it soon overflows the range of extended :- }
  powertable : array [0 .. 200] of extended;
  i : integer;
begin  { of TForm1.AllExtendedCalculations3Click }
  wr('Calculation of Pi using the amazing Chudnovsky Brothers formula');
  { Set up table for the term (640320^3)^n }
  powertable[0] := 1.0;
  for i := 1 to 200 do
    try
      powertable[i] := powertable[i-1]*const3;
    except
      ShowMessage('Failed on table fill at '+IntToStr(i));
      break
    end;
  term := 0.0;
  n := 0.0;
  flip := 1.0;
  wr('NOTE: Pi = 1/(12*Term)');
  repeat
    lastterm := term;
    f1 := fact(6.0*n) / (pwr3(fact(n)) * fact(3.0*n));
    f2 := (13591409.0 + n*545140134.0) / (powertable[round(n)]*sqrt(const3));
    term := term + flip*f1*f2;
    if flip = 1.0 then
      flip := -1.0
    else
      flip := 1.0;
    n := n + 1.0;
    wr('Term '+em(n,6)+#9#9+em(term,20))
  until lastterm = term;
  pi1 := 12.0*term;
  zpi := 1.0 / pi1;
  wr('Converged beyond limit of extended precision after only '+
     em(n,12)+' terms!');
  wr('Pi = '+em(zpi,20));
  wr('Difference between this value and the built-in "Pi" function = '+
     em(zpi-Pi,20));
  wr(' ');
end;  { of TForm1.AllExtendedCalculations3Click }

procedure TForm1.Improvedwithstats1Click(Sender: TObject);
{ This is an improved version of the Chudnovsky Brothers Pi formula, using
  integer type for "n", and adding up calculation stats. }
const
  maxterms = 1000;  { The maximum number of terms }
  numterms : integer = 100;  { The actual number of terms for this run }
var
  nadd, nsub, nmult, ndiv : integer;
{ Procedures to add up the basic arithmetic operations }
procedure na;
begin
  inc(nadd)
end;
procedure ns;
begin
  inc(nsub)
end;
procedure nm;
begin
  inc(nmult)
end;
procedure nd;
begin
  inc(ndiv)
end;
function fact (n : integer) : extended;
{ A local factorial function, where the argument "n" is integer }
var
  t : extended;
begin
  if n < 2 then
    fact := 1.0
  else
  begin
    t := n;
    while n > 1 do
    begin
      n := n - 1;
      t := t * n; nm;
    end;
    fact := t
  end
end;  { of fact }
function pwr3 (x : extended) : extended;
begin
  pwr3 := x*x*x; nm; nm;
end;
const
  const3 = 640320.0 * 640320.0 * 640320.0;
var
  n, flip : integer;
  term, f1, f2, pi1, zpi, lastterm, t : extended;
  { This is a table to do the calculation of "(640320^3)^n" }
  powertable : array [0 .. maxterms] of extended;
  i : integer;
  overflow, shown : boolean;
  sq : string;
begin  { of TForm1.Improvedwithstats1Click }
  wr('Calculation of Pi using the amazing Chudnovsky Brothers formula');
  sq := IntToStr(numterms);
  if InputQuery('Number of Terms', 'Give number of terms (currently '+
                em(numterms,12) + ')', sq) then
  begin
    numterms := StrToIntDef(sq, 10);
  end;
  if numterms > maxterms then
  begin
    numterms := maxterms;
    ShowMessage('Max is '+em(numterms,12));
  end;
  if numterms < 4 then
  begin
    numterms := 4;
    ShowMessage('Min is 4');
  end;
  nadd := 0;
  nsub := 0;
  nmult := 0;
  ndiv := 0;
  { Set up table for the term (640320^3)^n }
  powertable[0] := 1.0;
  for i := 1 to numterms do
    try
      powertable[i] := powertable[i-1]*const3; nm;
    except
      ShowMessage('Failed on table fill at '+IntToStr(i));
      break
    end;
  term := 0.0;
  n := 0;
  flip := 1;
  zpi := 0.0;
  wr('NOTE: Pi = 1/(12*Term)');
  overflow := false;
  shown := false;
  repeat
    try
      lastterm := term;
      f1 := fact(6*n) / (pwr3(fact(n)) * fact(3*n)); nm; nd; nm;
      f2 := (13591409.0 + n*545140134.0) / (powertable[round(n)]*sqrt(const3));
             na; nm; nd; nm;
      t := flip*f1*f2; nm; nm;
      term := term + t; na;
      if flip = 1 then
        flip := -1
      else
        flip := 1;
      inc (n);
      wr('Term '+em(n,6)+#9+em(term,20)+','+#9+' (Add '+em(t,20)+')');
      if lastterm = term then
      begin
        if not shown then
        begin
          wr('Converged beyond limit of extended precision after only '+
             em(n,12)+' terms!');
          shown := true;
        end;
      end;
    except
      wr('Reached arithmetic limit somewhere!');
      overflow := true;
    end;
  until overflow or (n >= numterms);
  try
    pi1 := 12.0*term;
    zpi := 1.0 / pi1;
  except
    wr('Reached arithmetic limit doing final calcs!');
  end;
  wr('Pi = '+em(zpi,20));
  wr('Difference between this value and the built-in "Pi" function = '+
     em(zpi-Pi,20));
  wr('Stats:');
  wr(' Add : '+em(nadd,9));
  wr(' Sub : '+em(nsub,9));
  wr(' Mult: '+em(nmult,9));
  wr(' Div : '+em(ndiv,9));
  wr(' ');
end;  { of TForm1.Improvedwithstats1Click }

procedure TForm1.AllExtendedCalculations4Click(Sender: TObject);
{ This procedure calculates Pi using the so-called "Algorithm 1". This is
  claimed to be the most efficient algorithm for Pi yet found! It is also
  the simplest by far! }
const
  maxterms = 100;
var
  a, y, a1, y1, lastterm, t, p2, zpi : extended;
  n : integer;
function pwr4(x : extended) : extended;
{ Returns x^4 }
begin
  pwr4 := x*x*x*x;
end;
begin
  wr('Calculation of Pi using the amazing so-called "Algorithm 1" formula');
  { p2 calculates the term "2^[2n+3]" by starting at 8, then simply executing
    "p2:=4p2" for the next term }
  p2 := 8.0;
  { Initial values for y[0] and a[0] according to algorithm :- }
  y := sqrt(2.0) - 1.0;
  a := 6.0 - (4.0*sqrt(2.0));
  n := 0;
  wr('NOTE: Pi = 1/a');
  repeat
    lastterm := a;
    wr('Term  '+em(n,12)+','+#9+'a = '+em(a,20)+','+#9+'y = '+
       em(y,20));
    { Next term values for y[n] and a[n] according to algorithm. Note that
      by the 4th term, y reaches 0 even with extended precision arithmetic!! }
    t := sqrt(sqrt(1.0-pwr4(y)));
    y1 := (1.0-t)/(1.0+t);
    a1 := pwr4(1.0+y1)*a - (p2*y1*(1.0 + y1 + y1*y1));
    p2 := p2*4.0;
    y := y1;
    a := a1;
    inc(n);
  until (n >= maxterms) or (lastterm = a);
  wr('Converged beyond limit of extended precision after only '+em(n,12)+
     ' terms!');
  wr('Term  '+em(n,12)+','+#9+'a = '+em(a,20)+','+#9+'y = '+
     em(y,20));
  zpi := 1.0/a;
  wr('Pi = '+em(zpi,20));
  wr('Difference between this value and the built-in "Pi" function = '+
     em(zpi-Pi,20));
  wr(' ');
end;

end.

